/* A SIMPLE PROGRAM TO CONVERT A CP/M-86 .H86 FILE TO CP/M-80 .HEX FORMAT 

   It is written because GENCMD.COM will not work with a .H86 file that has a
   final address above the BIOS in a CP/M-80 system. Also that nasty header in 
   a .CMD file is not present to mess ZSID up.  Note the program assumes the 
   program will run in a < 64k system.  The program was compiled with the AZTEC
   C compiler (Manx Software Systems).				     

		Author John Monahan			
	V1.1	28/2/10	Fixed up situations where H86 has multiple segments	*/


#include libc.h

main(argc,argv)
int argc;
char *argv[];
{
char *p, *d, *e, *lp, c, outfile[15], infile[15], info[257], cksum[5],line_buffer[256];
int k, i, fd, fd1, sum, type, new_sum,lp_count;

printf("\nThis program will convert a .H86 file written by the CP/M-86 V1.1");
printf("\nassembler (ASM86)to a file in CP/M-80 .HEX format.");
printf("\nNote only one 64K segment is assumed.\n");
printf("\n\Note CRT data dump has been inactivated. See source code to activate.\n");

if(argc != 2)
	{
	printf("\nUsage:-  H86TOHEX filename.H86\n");
	exit(0);
	}
for(p = argv[1]; *p; ++p)
	{
	*p = toupper(*p);
	if(*p == '?' || *p == '*' || *p < ' ' || *p == 0x7f)
		{
		printf("\nSorry cannot have ambigious file name!\n");
		exit(0);
		}
	}

p = argv[1];				/* name output file as .HEX */
d = outfile;
e = infile;

while(*p)
	{
	if(*p == '.')
		break;
	*d++ = *p;
	*e++ = *p++;
	}
strcpy(d,".HEX");
strcpy(e,".H86");
fd = fopen(infile,"r");
	if(fd == NULL)
		{
		printf("\nCannot open .H86 file");
		exit(0);
		}
;
fd1 = fopen(outfile,"w");
	if(fd1 == NULL)
		{
		printf("\nCannot open a .HEX file");
		exit(0);
		}
if(getc(fd) != ':')
		error(fd,fd1,"No starting ':' in .H86 file");

start:
getw(fd);					/* skip byte count */
getw(fd);					/* skip 0000 */
getw(fd);					
if((c = getc(fd)) == '0')			/* check for type 03 */
	{
	if(getc(fd) == '3')
		goto ok;
	}
else	{
	printf("\nSegment data etc. at start of .H86 file ignored\n");
	while ((c = getc(fd)) != ':')		/* ignore any segment types */
		{
		if(c == CPMEOF || c == PAD)
			error(fd,fd1,"Cannot find starting 03 record type");
		}
	goto start;
	}

ok:
printf("\n\nCode segment address = ");
for(i=0; i<4; i++)
	putchar(getc(fd));
printf("  Offset address = ");

for(i=0; i<4; i++)
	putchar(getc(fd));

getw(fd);					/* skip checksum */
if(agetc(fd) != '\n')		         
		error(fd,fd1,"No CR/LF after address header info");
putchar('\n');
putchar('\n');

/* --------------> Start of real data at this point <------------------------*/
	    /* program loops back to here for each line */

more:
if(getc(fd) != ':')
		error(fd,fd1,"No ':' info at start of line");
more1:
lp=line_buffer;					/* setup to fill a line */ 
*lp++ = ':';
/* putchar(':');  */

for(i=0; i<6; i++)				/* drop in byte count & address (6 bytes) */
	{
	c = getc(fd);
	*lp++ = c;
/*	if(i == 2)  */
/*		putchar(' ');	*/		/* space for easy reading */
/*	putchar(c);		*/
	}

p = info;					/* start of data store */
if((*p++ = getc(fd)) != '8')			/* Get type field */
	{					/* if 81,83... ok */
	if(getc(fd) == '1')			/* 01 if EOF */
		{
		putc(':',fd1);
		for(i=0; i<7; i++)		/* finish up hex file */
			putc('0',fd1);
/*		printf(" :00 0000 01 FF"); */	/* Last line = 00 0000 00 01FF */ 
		putc('1',fd1);
		putc('F',fd1);
		putc('F',fd1);
		goto finish;
		}
	else	error(fd,fd1,"Wrong record type in file");   
	}	

if((*p = getc(fd)) > '4') 			/* Next byte, is it 8[4..8] */
		{
		if(*p < '8')
		    {
   		    printf(" 8%c <-- Segment record type data ignored. Next line...\n",*p);
		    while(c = getc(fd) != ':');
		    goto more1;
		    }
		else	error(fd,fd1,"Wrong record type in file");
		}
p++;
*lp++ = '0';
*lp++ = '0';
/* printf(" 00 ");	  */				/* + spaces for easy reading */

while((*p = agetc(fd)) != ':')			/* get a line of hex code */
		{
		if(*p == CPMEOF || *p++ == 0x1a)
			error(fd,fd1,"\n\7Premature end of file detected!!\n");
		}
ungetc(':',fd);
						/* must now fix new checksum */

sum = hex_to_dec(p-3);				/* convert to decimal */
type = hex_to_dec(info);			/*need to correct this much */
new_sum = (sum + type);
sprintf(cksum,"%4x",new_sum);
*(p-3) = toupper(cksum[2]);			/* allow padding for > ffh */
*(p-2) = toupper(cksum[3]);

for(p = &info[2]; *p != ':'; p++)
	{
	*lp++ = *p;
/*	putchar(*p);  */ 
	}

*lp++ = 0;
lp=line_buffer;
while(*lp)
	aputc(*lp++,fd1);

goto more;

finish:	fclose(fd);
	fclose(fd1);
	printf("\n\nAll done\n");
	exit(0);
}				/* end of main program */

error(fd,fd1,string)
char *string;
int fd,fd1;
{
printf("\n%s\n",string);
fclose(fd);
fclose(fd1);
exit(0);
}

hex_to_dec(store)			/*this is a simple routine to convert*/
char store[4];				/*the two hex characters to decimal */
{
int temp,temp1;
char *s;
s = store;
if(*s > '9')
	temp1 = 9 + (*s - '@');
else temp1 = *s - '0';

if(*++s > '9')
	temp = 9 + (*s - '@');
else temp = *s - '0';

return(temp1*16 + temp);
}
